"""Base objects for definding templates

Born from a
`stackoverflow Q&A <https://stackoverflow.com/questions/69383734/templated
-object-generation-in-python/69384846>`_.


>>> from embody.templater import Templater
>>> # the following template has templated dicts, strings, and lists
>>> template = {
...     'hello': '{name}',
...     'how are you': ['{verb}', 2, '{name} and {verb} again']
... }
>>> g = Templater.template_func(template=template)
>>> g(name='NAME', verb="VERB")
{'hello': 'NAME', 'how are you': ['VERB', 2, 'NAME and VERB again']}
>>> str(g.__signature__)
'(*, name, verb)'
"""

import string
from functools import partial
from inspect import Signature, Parameter
from operator import itemgetter
from typing import Any, TypeVar, Tuple, Dict, List
from collections.abc import Callable, Generator, Iterable
from collections import namedtuple

T = TypeVar("T")
U = TypeVar("U")
K = TypeVar("K")
V = TypeVar("V")

# TODO: Make these things picklable


def get_generator_return(gen: Generator[T, Any, U]) -> tuple[Generator[T, Any, U], U]:
    return_value = None

    def inner():
        nonlocal return_value
        return_value = yield from gen

    gen_items = list(inner())

    def new_gen():
        yield from gen_items
        return return_value

    return new_gen(), return_value


def unique_list_conserving_order(iterable: Iterable) -> list:
    """List of unique elements in the original order
    >>> unique_list_conserving_order([3, 2, 3, 5, 1, 2, 6, 3, 5])
    [3, 2, 5, 1, 6]
    """
    seen = set()
    return [x for x in iterable if len(seen) < len(seen.add(x) or seen)]


# TemplateFunc: TypeAlias = Generator[str, None, Callable[..., T]]
TemplateFunc = Generator[str, None, Callable[..., T]]


class Templater:
    templater_registry: dict[type, Callable[[Any], TemplateFunc]] = {}

    @classmethod
    def register(cls, handles_type: type):
        def decorator(f):
            cls.templater_registry[handles_type] = f
            return f

        return decorator

    @classmethod
    def template_func_generator(cls, template: T) -> TemplateFunc[T]:
        if type(template) in cls.templater_registry:
            template_factory = cls.templater_registry[type(template)]
            return template_factory(template)
        else:
            # an empty generator that returns a function that returns the template unchanged,
            # since we don't know how to handle it
            def just_return():
                return lambda: template
                yield  # this yield is needed to tell python that this is a generator

            return just_return()

    @classmethod
    def template_func(cls, template: T) -> Callable[..., T]:
        gen = cls.template_func_generator(template)
        params, f = get_generator_return(gen)

        parameters = unique_list_conserving_order(
            Parameter(name=param, kind=Parameter.KEYWORD_ONLY) for param in params
        )
        f.__signature__ = Signature(parameters)
        return f


@Templater.register(str)
def templated_string_func(template: str) -> TemplateFunc[str]:
    """A function making templated strings. Like template.format, but with a signature"""
    f = partial(str.format, template)
    yield from filter(None, map(itemgetter(1), string.Formatter().parse(template)))

    return f


@Templater.register(dict)
def templated_dict_func(template: dict[K, V]) -> TemplateFunc[dict[K, V]]:
    DictEntryInfo = namedtuple(
        "DictEntryInfo", ["key_func", "value_func", "key_args", "value_args"]
    )
    entries: list[DictEntryInfo] = []
    for key, value in template.items():
        key_params, key_template_func = get_generator_return(
            Templater.template_func_generator(key)
        )
        value_params, value_template_func = get_generator_return(
            Templater.template_func_generator(value)
        )
        key_params = tuple(key_params)
        value_params = tuple(value_params)
        yield from key_params
        yield from value_params

        entries.append(
            DictEntryInfo(
                key_template_func, value_template_func, key_params, value_params
            )
        )

    def template_func(**kwargs):
        return {
            entry_info.key_func(
                **{arg: kwargs[arg] for arg in entry_info.key_args}
            ): entry_info.value_func(
                **{arg: kwargs[arg] for arg in entry_info.value_args}
            )
            for entry_info in entries
        }

    return template_func


@Templater.register(list)
def templated_list_func(template: list[T]) -> TemplateFunc[list[T]]:
    entries = []
    for item in template:
        params, item_template_func = get_generator_return(
            Templater.template_func_generator(item)
        )
        params = tuple(params)
        yield from params

        entries.append((item_template_func, params))

    # How is this used?
    def template_func(**kwargs):
        # print(kwargs)
        return [
            item_template_func(**{arg: kwargs[arg] for arg in args})
            for item_template_func, args in entries
        ]

    return template_func



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options.
# For a full list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------
from epythet.config_parser import parse_config
from pathlib import Path

project, copyright, author, release, display_name = parse_config(
    Path(__file__).absolute().parent.parent / "setup.cfg"
)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_toggleprompt",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",  # Include documentation from docstrings
    "sphinx.ext.doctest",  # Test snippets in the documentation
    "sphinx.ext.githubpages",  # This extension creates .nojekyll file
    "sphinx.ext.graphviz",  # Add Graphviz graphs
    "sphinx.ext.napoleon",  # Support for NumPy and Google style docstrings
    "sphinx.ext.todo",  # Support for todo items
    "sphinx.ext.viewcode",  # Add links to highlighted source code
    "myst_parser",  # Parse .md files
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for Markdown support -------------------------------------------
import commonmark


def docstring(app, what, name, obj, options, lines):
    md = "\n".join(lines)
    ast = commonmark.Parser().parse(md)
    rst = commonmark.ReStructuredTextRenderer().render(ast)
    lines.clear()
    lines += rst.splitlines()


def setup(app):
    app.connect("autodoc-process-docstring", docstring)


toggleprompt_offset_right = 25



"""Graphviz templates"""

from contextlib import suppress

with suppress(ImportError, ModuleNotFoundError):
    import graphviz

    def graphviz_attrs(**attrs):
        """
        >>> graphviz_attrs(resolution=23, label="My name")
        '[resolution=23, label="My name"]'
        """

        def gen():
            for k, v in attrs.items():
                if isinstance(v, str):
                    yield f'{k}="{v}"'
                else:
                    yield f"{k}={v}"

        return "[" + ", ".join(gen()) + "]"

    class graph_template:
        def __init__(
            self, name_attrs_str="[shape=rectangle style=filled color=lightgrey]"
        ):
            self.name_attrs_str = name_attrs_str

        def one_to_one(self, name="one_to_one", in_="in", out_="out", label="1-to-1"):
            return f'''
        subgraph cluster_{name} {{
            label = "{label}";
            {name}_in -> {name} -> {name}_out;
            {name}_in [label="{in_}"]
            {name}_out [label="{out_}"]
            {name} {self.name_attrs_str}
        }}'''

        def one_to_many(
            self,
            name="one_to_many",
            in_="in",
            out_1="out_1",
            out_n="out_n",
            label="1-to-many",
        ):
            return f'''
        subgraph cluster_{name} {{
            label = "{label}";
            {name}_in -> {name} -> {name}_out_1, {name}_out_n;
            {name}_in [label="{in_}"]
            {name}_out_1 [label="{out_1}"]
            {name}_out_n [label="{out_n}"]
            {name} {self.name_attrs_str}
        }}'''

        def many_to_one(
            self,
            name="many_to_one",
            in_1="in_1",
            in_n="in_n",
            out_="out",
            label="many-to-1",
        ):
            return f'''
        subgraph cluster_{name} {{
            label = "many-to-1";
            {name}_in_1, {name}_in_n -> {name} -> {name}_out;
            {name}_in_1 [label="{in_1}"]
            {name}_in_n [label="{in_n}"]
            {name}_out [label="{out_}"]
            {name} {self.name_attrs_str}
        }}'''

    dflt_graph_template = graph_template()

    def mk_graph_source(graph_template=dflt_graph_template, attrs=None, **name_kind):
        attrs = attrs or {}
        template = (
            "digraph G {{\n"
            + "graph {}\n".format(graphviz_attrs(**attrs.get("graph", {})))
            + "node {}\n".format(graphviz_attrs(**attrs.get("node", {})))
            + "edge {}\n".format(graphviz_attrs(**attrs.get("edge", {})))
            + "{}".format("\t\t{}\n" * len(name_kind))
            + "}}"
        )

        def format_specs():
            for name, kind in name_kind.items():
                if isinstance(kind, dict):
                    kind = dict(**kind)  # make a copy
                    method_name = kind.pop("kind")
                    method_kwargs = dict(name=name, **kind)
                    method = getattr(graph_template, method_name)
                    yield method(**method_kwargs)
                else:
                    if isinstance(kind, str):
                        method_name = kind
                        method_args = (name,)
                    elif isinstance(kind, tuple):
                        method_name, *method_args = kind
                    else:
                        raise TypeError("Can't recognize this kind of kind: {kind}")
                    method = getattr(graph_template, method_name)
                    yield method(*method_args)

        return template.format(*list(format_specs()))

    def mk_graph(graph_template=dflt_graph_template, **name_kind):
        return graphviz.Source(
            mk_graph_source(graph_template=graph_template, **name_kind)
        )

    def example():
        """You really need to use this code in a notebook, or capture"""
        import graphviz

        gt = graph_template()

        graph = [
            mk_graph(gt, function="one_to_one", wf_source="one_to_many"),
            mk_graph(
                gt, **{k: "one_to_many" for k in ("generator", "wf_source", "chunker")}
            ),
            mk_graph(
                gt,
                generator=dict(kind="one_to_many"),
                wf_source=dict(
                    kind="one_to_many", in_="source", out_1="wf_1", out_n="wf_n"
                ),
                chunker=dict(
                    kind="one_to_many", in_="wf", out_1="chk_1", out_n="chk_n"
                ),
            ),
            mk_graph(
                gt,
                function=dict(kind="one_to_one"),
                featurizer=dict(kind="one_to_one", in_="chk", out_="fv"),
                quantizer=dict(kind="one_to_one", in_="fv", out_="snip"),
                snip_stats=dict(kind="one_to_one", in_="snip", out_="stats"),
            ),
            mk_graph(
                gt,
                aggregator=dict(kind="many_to_one"),
                model_output=dict(
                    kind="many_to_one",
                    in_1="stats_1",
                    in_n="stats_n",
                    out_="detection_info",
                ),
            ),
            mk_graph(
                attrs={"graph": dict(size="9,9!"), "node": dict(size=3)},
                **{k: "one_to_many" for k in ("wf_source", "chunker")},
            ),
            mk_graph(
                attrs={
                    "graph": dict(size="6,6!", fontsize=20),
                    "node": dict(
                        shape="circle",
                        style="filled",
                        color="black",
                        fixedsize="true",
                        width="1.1",
                        height="1.1",
                        fontsize=15,
                    ),
                },
                **{k: "one_to_many" for k in ("wf_source", "chunker")},
            ),
        ]

        import os

        graph[3].render(filename=os.path.expanduser("~/example_graph.pdf"))
